/*
 * Intrinsics.cpp
 *
 *  Created on: Aug 10, 2015
 *      Author: Jan Vrany
 */

#ifndef LLVM_C_EXT_INTRINSICS_H
#define LLVM_C_EXT_INTRINSICS_H

#include <cstring>

#include <llvm/ADT/ArrayRef.h>
#include <llvm/IR/Intrinsics.h>
#include <llvm/IR/Module.h>
#include <llvm-c-ext/Intrinsics.h>


static const char* const intrinsic_names[] = {
#define GET_INTRINSIC_NAME_TABLE
#include "llvm/IR/Intrinsics.gen"
#undef GET_INTRINSIC_NAME_TABLE
};

using namespace llvm;

LLVMValueRef LLVMGetIntrinsicByName(LLVMModuleRef mod, const char* name, const int count, const LLVMTypeRef* types) {
	Function* f;
    Type* typesUnwrapped[32];

    assert( (count <= 32) && "Intrinsics with more than 32 arguments are not supported");

    for (int typeIdx = 0; typeIdx < count; typeIdx++) {
        typesUnwrapped[typeIdx] = unwrap(types[typeIdx]);
    }
    ArrayRef<Type*> typesArray(typesUnwrapped, count);
	for (unsigned int intrinIdx = 0; intrinIdx < (sizeof(intrinsic_names) / sizeof(const char*)); intrinIdx++) {
		if (strcmp(name, intrinsic_names[intrinIdx]) == 0) {
			
			f = Intrinsic::getDeclaration(unwrap(mod), (Intrinsic::ID)(intrinIdx + 1), typesArray);
			return wrap(f);
		}
	}
	return NULL;
}

#endif /* !defined(LLVM_C_EXT_INTRINSICS_H) */








